from _decimal import Decimal
from datetime import date, datetime, time
from logging import getLogger
from typing import Dict, Iterable, List, Optional, Tuple, Union

from NEMO.models import Account, Project
from NEMO.utilities import get_month_timeframe, localize
from dateutil.relativedelta import relativedelta
from django.contrib import messages
from django.contrib.admin.utils import get_deleted_objects
from django.db import transaction
from django.db.models import Count, Q
from django.http import HttpRequest
from django.utils import timezone

from NEMO_billing.cap_discount.customization import CAPDiscountCustomization
from NEMO_billing.cap_discount.exceptions import MissingCAPAmountException, NotLatestInvoiceException
from NEMO_billing.cap_discount.models import CAPDiscount, CAPDiscountAmount, CAPDiscountConfiguration, CAPDiscountTier
from NEMO_billing.invoices.customization import InvoiceCustomization
from NEMO_billing.invoices.models import (
    BillableItemType,
    Invoice,
    InvoiceConfiguration,
    InvoiceDetailItem,
    InvoiceSummaryItem,
)
from NEMO_billing.invoices.processors import (
    FullInvoice,
    InvoiceDataProcessor,
    billable_to_invoice_detail_item,
    get_account_invoices,
)
from NEMO_billing.rates.models import RateCategory
from NEMO_billing.utilities import round_decimal_amount

processors_logger = getLogger(__name__)


error_extra_tags = "data-speed=30000 data-trigger=manual"


# Class to hold information about the discounts for an account, split by projects/users/core facilities
class AccountDiscountCalculator:
    def __init__(self, start_date: date):
        self.start_date = start_date
        # We are creating a dict of rate_category -> tuple(facility_names) -> usernames -> cap discount objects
        self.cap_discounts: Dict[str, Dict[tuple, Dict[str, CAPDiscount]]] = {}

    def add_item(self, item: InvoiceDetailItem):
        # This assumes we are getting the items in order of end date first, so we can add up amounts
        project = item.invoice.project_details.project
        cap_discount = self.get_or_create_cap(project, item)
        if cap_discount and item_cap_discount_match(item, cap_discount):
            discount_amount = self.calculate_discount(cap_discount.tmp_current_amount, item.amount, cap_discount)
            if discount_amount:
                prj_id = item.invoice.project_details.project.id
                project_facility_discount = cap_discount.tmp_project_discounts.setdefault(prj_id, {}).setdefault(
                    item.core_facility, Decimal(0)
                )
                item.discount = discount_amount
                cap_discount.tmp_project_discounts[prj_id][item.core_facility] = (
                    project_facility_discount + item.discount
                )
            cap_discount.tmp_current_amount += item.amount

    def calculate_discount(self, current_total, item_amount, cap_discount: CAPDiscount) -> Decimal:
        total_discount = Decimal(0)
        # We need to order by descending tier amount. This is really important for the algorithm to work
        tier_discounts = list(cap_discount.capdiscounttier_set.order_by("-amount"))
        # We are initializing the remaining at the current amount
        # Then we will apply the higher tier discount if applicable,
        # And subtract the amount until we have nothing remaining
        remaining = item_amount
        for i, tier in enumerate(tier_discounts):
            tier: CAPDiscountTier = tier
            if remaining and current_total + remaining > tier.amount:
                amount_for_discount = remaining - max(tier.amount - current_total, Decimal(0))
                total_discount += amount_for_discount * tier.discount_amount()
                remaining -= amount_for_discount
        return round_decimal_amount(total_discount)

    def get_or_create_cap(self, project: Project, item: InvoiceDetailItem) -> Optional[CAPDiscount]:
        core_facility = item.core_facility
        username = item.user
        rate_category = project.projectbillingdetails.category
        rate_category_dict: Dict[tuple, Dict[str, CAPDiscount]] = self.cap_discounts.setdefault(rate_category.name, {})
        try:
            config = CAPDiscountConfiguration.objects.get(
                rate_category=rate_category, core_facilities__name=core_facility
            )
            facility_names = config.core_facility_names or (None,)
            facility_dict: Dict[str, CAPDiscount] = rate_category_dict.setdefault(facility_names, {})
            username = username if config.split_by_user else None
            if username in facility_dict:
                return facility_dict.get(username)
            # Only create if it is an eligible charge
            elif item_cap_discount_match(item, config):
                cap_discount: CAPDiscount = config.get_or_create_cap_discount(
                    account=project.account, username=username, start=self.start_date
                )
                # Set tmp current amount
                cap_discount.tmp_current_amount = CAPDiscountAmount.objects.get(
                    cap_discount=cap_discount, month=self.start_date.month, year=self.start_date.year
                ).start
                return facility_dict.setdefault(username, cap_discount)
        except CAPDiscountConfiguration.DoesNotExist:
            processors_logger.debug(
                f"No CAP configuration found for rate: {rate_category} and core facility: {core_facility}"
            )

    def get_discounts_for_project(self, rate_category: RateCategory, core_facility: str) -> List[CAPDiscount]:
        # Returns the facility discounts for each user
        total_cap_discounts: List[CAPDiscount] = []
        for core_facility_names, cap_discounts_dict in self.cap_discounts.get(rate_category.name, {}).items():
            if core_facility in core_facility_names:
                total_cap_discounts.extend(cap_discounts_dict.values())
        return total_cap_discounts

    def all_cap_discounts(self) -> List[CAPDiscount]:
        return [
            value
            for rate_category_dict in self.cap_discounts.values()
            for facility_dict in rate_category_dict.values()
            for value in facility_dict.values()
            if rate_category_dict and facility_dict
        ]


class CAPDiscountInvoiceDataProcessor(InvoiceDataProcessor):
    # Do all this in a transaction so everything gets rolled back if an error happened
    @transaction.atomic
    def generate_invoice_for_account(self, month, account, configuration, user, raise_if_exists=False) -> List[Invoice]:
        capped_invoices: List[FullInvoice] = []
        non_capped_invoices: List[FullInvoice] = []
        if not InvoiceCustomization.get_bool("invoice_skip_inactive_accounts") or account.active:
            # Overriding this, so we can deal with discounts and save all invoices at the end
            start, end = get_month_timeframe(month)
            # First let's check that cap discounts have been correctly generated until now
            # And initialize new amounts for this month
            check_and_initialize_monthly_cap_amounts(account, start)

            # Now create our cap calculator helper
            discount_calculator = AccountDiscountCalculator(start.date())

            # Go through the generation process, and add all items to account details list
            all_account_details: List[InvoiceDetailItem] = []
            account_filter = Q(project__account_id=account.id)
            all_billable_items = self.get_billable_items(
                start, end, configuration, account_filter, account_filter, account_filter
            )
            projects = account.project_set.all()
            if InvoiceCustomization.get_bool("invoice_skip_inactive_projects"):
                projects = projects.filter(active=True)
            for project in projects:
                billables = [billable for billable in all_billable_items if billable.project == project]
                invoice = self.generate_invoice(start, end, project, configuration, user, billables, raise_if_exists)
                if invoice:
                    # do not run cap calculations for prepaid projects
                    if hasattr(project, "projectprepaymentdetail"):
                        non_capped_invoices.append(invoice)
                    else:
                        capped_invoices.append(invoice)
                        all_account_details.extend(invoice.detail_items)

            if all_account_details:
                # Sort account details by end date. This is very IMPORTANT to get discounts right
                all_account_details.sort(key=lambda x: x.end)
                # Go through all account charges to figure out the discounts
                for detail in all_account_details:
                    discount_calculator.add_item(detail)
                # Now let's update the CAP entities with the total charges for this account
                for cap_discount in discount_calculator.all_cap_discounts():
                    amount = CAPDiscountAmount.objects.get(
                        cap_discount=cap_discount, year=start.year, month=start.month
                    )
                    amount.end = cap_discount.tmp_current_amount
                    amount.save(update_fields=["end"])

            # Save all the invoices now
            self.save_invoices(capped_invoices, discount_calculator)
            # Only save regular invoices after everything went well for CAP invoices
            for invoice in non_capped_invoices:
                invoice.save()
        return [full_invoice.invoice for full_invoice in (capped_invoices + non_capped_invoices)]

    def save_invoices(self, full_invoices: List[FullInvoice], discount_calculator: AccountDiscountCalculator = None):
        for full_invoice in full_invoices:
            # Pass calculator through the invoice, so we can retrieve it later
            full_invoice.invoice.discount_calculator = discount_calculator
            # Recalculate summaries with the discounts
            full_invoice.summary_items = self.get_invoice_summary_items(full_invoice.invoice, full_invoice.detail_items)
            full_invoice.save()

    # Override this method, to check if we need to deal with discounts or not
    def get_summary_items_for_facility(self, invoice: Invoice, core_facility: str, details) -> List[InvoiceSummaryItem]:
        cap_discounts = []

        calc: AccountDiscountCalculator = getattr(invoice, "discount_calculator", None)
        if calc:
            rate_category = invoice.project_details.category
            cap_discounts: List[CAPDiscount] = calc.get_discounts_for_project(rate_category, core_facility)

        if not cap_discounts:
            # If we don't have any discounts, return original function
            return super().get_summary_items_for_facility(invoice, core_facility, details)
        else:
            # Otherwise re-generate summary items with discounts
            return self.get_summary_items_for_facility_with_discounts(invoice, core_facility, details, cap_discounts)

    # This is our new version with the discount calculation
    def get_summary_items_for_facility_with_discounts(
        self, invoice: Invoice, core_facility: str, details, cap_discounts: List[CAPDiscount]
    ) -> List[InvoiceSummaryItem]:
        # Re-generate summary items
        facility_summaries: List[InvoiceSummaryItem] = []
        for billable_type in BillableItemType:
            items = [item for item in details if item.item_type == billable_type.value]
            facility_summaries.extend(self.get_recap_usage_summary(invoice, core_facility, items))

        total_discount = Decimal(0)
        for cap_discount in cap_discounts:
            discount_amt = cap_discount.tmp_project_discounts.get(invoice.project_details.project.id, {}).get(
                core_facility
            )
            if discount_amt and discount_amt != 0:
                discount_name = f"{cap_discount.configuration.rate_category.name}{'/' + cap_discount.user.username if cap_discount.user else ''} CAP Discount"
                f_discount = InvoiceSummaryItem(invoice=invoice, name=discount_name, core_facility=core_facility)
                f_discount.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.DISCOUNT
                f_discount.amount = round_decimal_amount(discount_amt)
                facility_summaries.append(f_discount)
                total_discount += f_discount.amount

        facility_subtotal = InvoiceSummaryItem(invoice=invoice, name="Subtotal", core_facility=core_facility)
        facility_subtotal.summary_item_type = InvoiceSummaryItem.InvoiceSummaryItemType.SUBTOTAL
        non_waived_items = [item for item in details if not item.waived]
        facility_subtotal.amount = round_decimal_amount(sum(item.amount for item in non_waived_items) + total_discount)
        facility_summaries.append(facility_subtotal)

        return facility_summaries

    @transaction.atomic
    def delete_invoice(self, invoice: Invoice, request: HttpRequest):
        # NOTE: For now, we are deleting all invoices for this account, even non-capped ones
        # if this invoice has a CAP Discount associated with its account,
        # delete all the other account invoices and CAP amounts for the month.
        # This will force to re-generate all the invoices and CAP calculation.
        invoice_account = invoice.project_details.project.account
        invoice_year, invoice_month = invoice.start.year, invoice.start.month
        try:
            delete_associated_amounts(invoice_account, invoice_year, invoice_month, request)
        except NotLatestInvoiceException as e:
            messages.error(request, e.msg, extra_tags=error_extra_tags)
            return
        # the super method will delete all other account invoices, so we are good
        super().delete_invoice(invoice, request)

    @transaction.atomic
    def void_invoice(self, invoice: Invoice, request: HttpRequest):
        invoice_account = invoice.project_details.project.account
        invoice_year, invoice_month = invoice.start.year, invoice.start.month
        try:
            delete_associated_amounts(invoice_account, invoice_year, invoice_month, request)
        except NotLatestInvoiceException as e:
            messages.error(request, e.msg, extra_tags=error_extra_tags)
            return
        # the super method will void all other account invoices, so we are good
        super().void_invoice(invoice, request)

    def get_deleted_objects(self, invoices: Iterable[Invoice], request: HttpRequest, admin_site):
        # gather all related objects to delete: other invoices, cap discounts, cap discount amounts
        related_invoices = []
        related_discounts = []
        related_amounts = []
        try:
            for invoice in invoices:
                invoice_account = invoice.project_details.project.account
                invoice_year, invoice_month = invoice.start.year, invoice.start.month
                related_invoices.extend(get_account_invoices(invoice_account, invoice_year, invoice_month))
                related_amounts.extend(get_amounts_to_delete(invoice_account, invoice_year, invoice_month))
        except NotLatestInvoiceException as e:
            return [], {}, set(), [e.msg]
        invoice_deleted_objects = get_deleted_objects(related_invoices, request, admin_site)
        # We might need to also delete some CAPDiscounts if the amount was the first
        for cap_amount in related_amounts:
            if cap_amount.cap_discount.earliest_amount() == cap_amount:
                related_discounts.append(cap_amount.cap_discount)
                related_amounts.remove(cap_amount)
        amount_deleted_objects = get_deleted_objects(related_amounts, request, admin_site)
        discount_deleted_objects = get_deleted_objects(related_discounts, request, admin_site)
        return (
            invoice_deleted_objects[0] + amount_deleted_objects[0] + discount_deleted_objects[0],
            {**invoice_deleted_objects[1], **amount_deleted_objects[1], **discount_deleted_objects[1]},
            invoice_deleted_objects[2],
            invoice_deleted_objects[2],
        )

    def estimated_charges_since_last_update(self, cap_discount: CAPDiscount) -> Decimal:
        latest_amount = cap_discount.latest_amount()
        start_since_update = (
            latest_amount.amount_date + relativedelta(months=1)
            if latest_amount
            else date(timezone.now().year, timezone.now().month, 1)
        )
        customer_filter = Q(project__account=cap_discount.account)
        user_filter = Q(project__account=cap_discount.account)
        trainee_filter = Q(project__account=cap_discount.account)
        if cap_discount.user:
            customer_filter = customer_filter & Q(customer=cap_discount.user)
            user_filter = user_filter & Q(user=cap_discount.user)
            trainee_filter = trainee_filter & Q(trainee=cap_discount.user)
        config = InvoiceConfiguration.first_or_default()
        items = self.get_billable_items(
            localize(datetime.combine(start_since_update, time())),
            timezone.now(),
            config,
            customer_filter,
            user_filter,
            trainee_filter,
            raise_no_rate=False,
        )
        detail_items: List[InvoiceDetailItem] = []
        for item in items:
            # exceptional case here where we need to create a "fake" invoice,
            # so we can match items correctly
            invoice = Invoice()
            invoice.project_details = item.project.projectbillingdetails
            detail_items.append(billable_to_invoice_detail_item(item, invoice=invoice))
        return Decimal(sum(item.amount for item in detail_items if item_cap_discount_match(item, cap_discount)))


def item_cap_discount_match(item: InvoiceDetailItem, cap: Union[CAPDiscount, CAPDiscountConfiguration]):
    # function to check eligible/matching items for CAPDiscount or CAPDiscountConfiguration
    configuration = cap
    cap_match = True
    if isinstance(cap, CAPDiscount):
        cap_match = item.invoice.project_details.project.account == cap.account
        if cap.configuration.split_by_user:
            cap_match = cap_match and item.user == cap.user.username
        configuration = cap.configuration

    configuration_core_facilities: Tuple[str] = configuration.core_facility_names or (None,)
    item_type = BillableItemType(item.item_type)

    # Check if tool is not excluded from CAP
    tool_excluded = False
    excluded_tool_ids = CAPDiscountCustomization.get_excluded_tool_ids()
    if excluded_tool_ids:
        if item_type in [BillableItemType.TOOL_USAGE, BillableItemType.MISSED_RESERVATION, BillableItemType.TRAINING]:
            if item.content_object and item.content_object.tool:
                tool_excluded = item.content_object.tool.id in excluded_tool_ids
    if item_type == BillableItemType.STAFF_CHARGE:
        # TODO: not sure here how to figure out if the tool is excluded since it's not set on the item
        pass

    # Only include cap eligible Custom Charges
    custom_charge_included = True
    if item_type == BillableItemType.CUSTOM_CHARGE:
        custom_charge_included = item.content_object and item.content_object.cap_eligible

    # Check if the project itself is included
    project_included = not item.invoice.project_details.no_cap

    return (
        item_type in cap.billable_charge_types
        and not item.waived
        and not tool_excluded
        and custom_charge_included
        and project_included
        and item.invoice.project_details.category == configuration.rate_category
        and item.core_facility in configuration_core_facilities
        and cap_match
    )


def check_and_initialize_monthly_cap_amounts(account: Account, start: date):
    # Check that current discount amounts for the account have a previous value, otherwise raise an exception.
    # We need to exclude the cap discounts that are part of the month we are generating the invoice for.
    # To avoid throwing errors for brand-new CAP when generating invoices multiple times over.
    caps = (
        CAPDiscount.objects.filter(account=account)
        .annotate(
            total_amounts=Count("capdiscountamount"),
            matching_amounts=Count(
                "capdiscountamount", filter=Q(capdiscountamount__year=start.year, capdiscountamount__month=start.month)
            ),
        )
        .exclude(total_amounts=1, matching_amounts=1)
        .values_list("id", flat=True)
    )
    previous_month = start - relativedelta(months=1)
    prev_cap_amounts = CAPDiscountAmount.objects.filter(
        cap_discount__in=caps, year=previous_month.year, month=previous_month.month
    )
    if prev_cap_amounts.count() != caps.count():
        raise MissingCAPAmountException(previous_month, account)

    # Initialize all discounts for this account
    for prev_cap_amount in prev_cap_amounts:
        initialize_cap_discount_amount(prev_cap_amount, start)


def initialize_cap_discount_amount(previous_amount: CAPDiscountAmount, start_date: date):
    cap_discount = previous_amount.cap_discount
    # Check if we need to reset the CAP
    start_for_rec = datetime(year=start_date.year, month=start_date.month, day=start_date.day)
    cap_reset = cap_discount.next_reset_date_after(start_date, inc=True)
    amount = Decimal(0)
    if start_for_rec == cap_reset:
        cap_discount.reset()
    # Only check previous amount if it's not a reset. Otherwise, it won't start over at 0
    elif previous_amount:
        amount = previous_amount.end
    discount_amount, created = CAPDiscountAmount.objects.get_or_create(
        cap_discount=cap_discount,
        year=start_date.year,
        month=start_date.month,
        defaults={"start": amount, "end": amount},
    )
    cap_discount.tmp_current_amount = discount_amount.start


def get_amounts_to_delete(account: Account, year, month) -> List[CAPDiscountAmount]:
    cap_amounts_to_delete: List[CAPDiscountAmount] = []
    # check to make sure they are all the latest invoices/amounts
    for cap_discount in CAPDiscount.objects.filter(account=account):
        latest_amount = cap_discount.latest_amount()
        if not latest_amount or not latest_amount.year == year or not latest_amount.month == month:
            raise NotLatestInvoiceException(latest_amount)
        cap_amounts_to_delete.append(latest_amount)
    return cap_amounts_to_delete


def delete_associated_amounts(account: Account, year, month, request):
    # delete associated CAP amounts for the year/month
    for cap_amount in get_amounts_to_delete(account, year, month):
        # If the amount was the first one, then delete the whole CAP discount
        if cap_amount.cap_discount.earliest_amount() == cap_amount:
            cap_amount.cap_discount.delete()
            messages.success(request, f"{cap_amount.cap_discount} was successfully deleted.", "data-speed=30000")
        else:
            cap_amount.delete()
            messages.success(request, f"{cap_amount} was successfully deleted.", "data-speed=30000")



# Generated by Django 3.2.16 on 2023-01-21 16:57

import re

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("NEMO_billing", "0002_version_1_2_0"),
        ("rates", "0003_version_1_17_0"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("NEMO", "0042_version_4_3_0"),
    ]

    operations = [
        migrations.CreateModel(
            name="CAPDiscountConfiguration",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Amount after which the discount will be applied",
                        max_digits=14,
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "discount",
                    models.DecimalField(
                        decimal_places=3,
                        help_text="Discount in percent. Ex 20.5%",
                        max_digits=6,
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "charge_types",
                    models.CharField(
                        help_text="List of charge type this CAP applies to",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^\\d+(?:,\\d+)*\\Z"),
                                code="invalid",
                                message="Enter only digits separated by commas.",
                            )
                        ],
                    ),
                ),
                (
                    "start_on_first_charge",
                    models.BooleanField(
                        default=True,
                        help_text="Automatically set CAP start date to the month of the first eligible charge",
                    ),
                ),
                (
                    "start_month",
                    models.PositiveIntegerField(
                        blank=True,
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ],
                        help_text="Start month of the CAP discount",
                        null=True,
                    ),
                ),
                (
                    "start_year",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Start year of the CAP discount",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1900),
                            django.core.validators.MaxValueValidator(9999),
                        ],
                    ),
                ),
                (
                    "reset_interval",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="Recurring interval, i.e. every 5 days.",
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "reset_frequency",
                    models.PositiveIntegerField(
                        choices=[(5, "Month(s)"), (6, "Year(s)")], help_text="The CAP reset frequency."
                    ),
                ),
                (
                    "core_facility",
                    models.ForeignKey(
                        blank=True,
                        help_text="The core facility this CAP applies to",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="NEMO_billing.corefacility",
                    ),
                ),
                (
                    "rate_category",
                    models.ForeignKey(
                        help_text="The rate category this CAP applies to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="rates.ratecategory",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CAPDiscount",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Amount after which the discount will be applied",
                        max_digits=14,
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "discount",
                    models.DecimalField(
                        decimal_places=3,
                        help_text="Discount in percent. Ex 20.5%",
                        max_digits=6,
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "charge_types",
                    models.CharField(
                        help_text="List of charge type this CAP applies to",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^\\d+(?:,\\d+)*\\Z"),
                                code="invalid",
                                message="Enter only digits separated by commas.",
                            )
                        ],
                    ),
                ),
                (
                    "start_month",
                    models.PositiveIntegerField(
                        blank=True,
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ],
                        help_text="Start month of the CAP discount",
                        null=True,
                    ),
                ),
                (
                    "start_year",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Start year of the CAP discount",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1900),
                            django.core.validators.MaxValueValidator(9999),
                        ],
                    ),
                ),
                (
                    "next_reset_override_month",
                    models.PositiveIntegerField(
                        blank=True,
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ],
                        help_text="Temporary one-time reset override month",
                        null=True,
                    ),
                ),
                (
                    "next_reset_override_year",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Temporary one-time reset override year",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1900),
                            django.core.validators.MaxValueValidator(9999),
                        ],
                    ),
                ),
                (
                    "reset_interval",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="Recurring interval, i.e. every 5 days.",
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "reset_frequency",
                    models.PositiveIntegerField(
                        choices=[(5, "Month(s)"), (6, "Year(s)")], help_text="The CAP reset frequency."
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        help_text="The account this CAP applies to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="NEMO.account",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="The user this CAP applies to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "configuration",
                    models.ForeignKey(
                        help_text="The configuration for this CAP",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cap_discount.capdiscountconfiguration",
                    ),
                ),
            ],
            options={
                "unique_together": {("configuration", "user", "account")},
            },
        ),
        migrations.CreateModel(
            name="CAPDiscountAmount",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "month",
                    models.PositiveIntegerField(
                        choices=[
                            (1, "JANUARY"),
                            (2, "FEBRUARY"),
                            (3, "MARCH"),
                            (4, "APRIL"),
                            (5, "MAY"),
                            (6, "JUNE"),
                            (7, "JULY"),
                            (8, "AUGUST"),
                            (9, "SEPTEMBER"),
                            (10, "OCTOBER"),
                            (11, "NOVEMBER"),
                            (12, "DECEMBER"),
                        ]
                    ),
                ),
                (
                    "year",
                    models.PositiveIntegerField(
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(9999),
                        ]
                    ),
                ),
                ("start", models.DecimalField(decimal_places=2, max_digits=14)),
                ("end", models.DecimalField(decimal_places=2, max_digits=14)),
                (
                    "cap_discount",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="cap_discount.capdiscount"),
                ),
            ],
            options={
                "ordering": ["-year", "-month"],
                "unique_together": {("cap_discount", "month", "year")},
            },
        ),
        migrations.AddConstraint(
            model_name="capdiscountconfiguration",
            constraint=models.UniqueConstraint(
                fields=("rate_category", "core_facility"), name="capdiscountconfiguration_unique_rate_category_facility"
            ),
        ),
        migrations.AddConstraint(
            model_name="capdiscountconfiguration",
            constraint=models.UniqueConstraint(
                condition=models.Q(("core_facility", None)),
                fields=("rate_category",),
                name="capdiscountconfiguration_unique_rate_category_null_facility",
            ),
        ),
    ]



from datetime import datetime
from logging import getLogger
from typing import Optional
from urllib.parse import urlencode

from NEMO.models import Area, Consumable, Project, Tool, User
from NEMO.utilities import render_email_template, send_mail
from NEMO.views.customization import get_media_file_contents
from NEMO.views.pagination import SortedPaginator
from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.http import FileResponse, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.http import require_GET, require_http_methods

from NEMO_billing.invoices.utilities import flatten
from NEMO_billing.quotes.exceptions import FailedToSendQuoteEmailException, QuoteGenerationException
from NEMO_billing.quotes.forms import CreateQuoteForm, EditQuoteForm, QuoteItemForm
from NEMO_billing.quotes.models import Quote, QuoteConfiguration, QuoteItem
from NEMO_billing.quotes.utilities import (
    QUOTE_EMAIL_CATEGORY,
    available_configurations_for_user_to_create,
    available_configurations_for_user_to_create_or_approve,
    can_access_quotes,
    can_create_quotes,
    create_approval_request_notification,
    delete_approval_request_notification,
    get_approval_request_notifications_per_configuration,
    meets_send_email_prerequisites,
)
from NEMO_billing.rates.models import Rate, RateType
from NEMO_billing.rates.utilities import RateHistory
from NEMO_billing.rates.views import get_rate_type_choices, get_rate_types

quotes_logger = getLogger(__name__)


@login_required
@user_passes_test(can_access_quotes)
@require_GET
def quote_list(request, selected_configuration_id=None):
    user = request.user
    quote_configurations = QuoteConfiguration.objects.all()
    configurations_info = []
    configurations_exist = quote_configurations.exists()
    page = []
    selected_configuration = None
    status_filter = request.GET.get("status")

    if configurations_exist:
        if not selected_configuration_id:
            user_configurations = available_configurations_for_user_to_create_or_approve(user)
            selected_configuration = user_configurations[0] if user_configurations else quote_configurations.first()
            if selected_configuration:
                return redirect("quotes", selected_configuration_id=selected_configuration.id)
        else:
            selected_configuration = get_object_or_404(QuoteConfiguration, pk=selected_configuration_id)

        # Filter quotes based on selected configuration and status filter and paginate results
        quotes = Quote.objects.all()
        if selected_configuration:
            quotes = quotes.filter(configuration=selected_configuration.id)
        if status_filter:
            quotes = quotes.filter(status=status_filter)
        page = SortedPaginator(quotes, request).get_current_page()

        # Fetch notifications for configurations
        notification_map = get_approval_request_notifications_per_configuration(user)
        for configuration in quote_configurations:
            configuration_info = {
                "id": configuration.id,
                "name": configuration.name,
                "notification_count": notification_map.get(configuration.id, 0),
            }
            configurations_info.append(configuration_info)

    return render(
        request,
        "quotes/quote_list.html",
        {
            "page": page,
            "quote_configurations": configurations_info,
            "configurations_exist": configurations_exist,
            "selected_configuration": selected_configuration,
            "quote_statuses": Quote.Status.Choices,
            "selected_status": status_filter,
            "user_can_create": selected_configuration.can_user_create(user) if selected_configuration else False,
        },
    )


@login_required
@user_passes_test(can_create_quotes)
@require_http_methods(["GET", "POST"])
def create_quote(request, selected_configuration_id=None):
    user = request.user
    available_configurations = available_configurations_for_user_to_create(user)
    selected_configuration = None

    if selected_configuration_id:
        selected_configuration = get_object_or_404(QuoteConfiguration, pk=selected_configuration_id)
        if selected_configuration not in available_configurations:
            messages.error(request, "You do not have permission to create quotes with the selected configuration.")
            return redirect("create_quote")

    form = CreateQuoteForm(request.POST or None)
    if request.method == "POST":
        if form.is_valid():
            try:
                # Check that the user has permission to create quotes with the selected configuration
                configuration = form.cleaned_data.get("configuration")
                if not configuration.can_user_create(user):
                    messages.error(
                        request, "You don't have permission to create quotes with the selected configuration."
                    )
                    return render(
                        request, "quotes/create_quote.html", {"form": form, "configurations": available_configurations}
                    )

                quote = form.save(commit=False)
                quote.creator = user
                quote.status = Quote.Status.DRAFT
                quote.save()

                messages.success(request, "Quote created successfully.")
                return redirect("quote", quote_id=quote.id)
            except QuoteConfiguration.DoesNotExist:
                messages.error(request, "Selected configuration does not exist.")

    return render(
        request,
        "quotes/create_quote.html",
        {
            "form": form,
            "configurations": available_configurations,
            "selected_configuration": selected_configuration,
            "user_can_create": selected_configuration.can_user_create(user),
        },
    )


@login_required
@user_passes_test(can_access_quotes)
@require_http_methods(["GET", "POST"])
def view_quote(request, quote_id):
    user = request.user
    rate_type_choice = request.GET.get("type")
    item_id = request.GET.get("id")
    quote_obj = get_object_or_404(Quote, id=quote_id)
    quote_items = quote_obj.quoteitem_set.all()
    form = EditQuoteForm(request.POST or None, instance=get_object_or_404(Quote, id=quote_id))
    permissions = quote_obj.permissions(user)
    submitted_users = None
    show_send_emails_button = meets_send_email_prerequisites(quote_obj) and permissions["can_send_emails"]
    active_quote_item_tab = "rate"

    if request.method == "POST":
        user_ids = form.data.getlist("users")
        submitted_users = User.objects.filter(id__in=user_ids)

        if not permissions["can_edit_metadata"]:
            messages.error(request, "Cannot edit quote.")
            return redirect("quote", quote_id=quote_id)

        if form.is_valid():
            quote_obj = form.save()
            # Re-render the quote file if it has been updated (the linked project might have changed)
            try:
                quote_obj.render_and_save()
            except QuoteGenerationException as e:
                messages.warning(request, f"Quote '{quote_obj.name}' updated successfully but failed to render file.")
                return redirect("quote", quote_id=quote_obj.id)

            messages.success(request, f"Quote '{quote_obj.name}' updated successfully!")
            return redirect("quote", quote_id=quote_obj.id)

    context = {
        "quote": quote_obj,
        "quote_items": quote_items,
        "submitted": quote_obj.status == Quote.Status.PENDING_APPROVAL,
        "published": quote_obj.status == Quote.Status.PUBLISHED,
        "draft": quote_obj.status == Quote.Status.DRAFT,
        "updated": quote_obj.updated_date != quote_obj.created_date,
        "users": User.objects.filter(is_active=True).exclude(id=user.id),
        "projects": Project.objects.all(),
        "form": form,
        "submitted_users": submitted_users,
        "active_quote_item_tab": active_quote_item_tab,
        "show_send_emails_button": show_send_emails_button,
    }
    rate_search_context = get_rate_search_context(rate_type_choice, item_id)
    context.update(rate_search_context)
    context.update(permissions)

    return render(request, "quotes/quote.html", context)


@login_required
@user_passes_test(can_create_quotes)
@require_http_methods(["POST"])
def add_quote_item(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    rate_type_choice = request.POST.get("rate_type_choice")
    item_id = request.POST.get("rate_type_item_id")
    permissions = quote_obj.permissions(user)
    show_send_emails_button = meets_send_email_prerequisites(quote_obj) and permissions["can_send_emails"]
    form = EditQuoteForm(instance=get_object_or_404(Quote, id=quote_id))
    # quote_item_type is used to determine which form was submitted (rate or custom)
    active_form_type = request.POST.get("quote_item_type")

    if not permissions["can_edit"]:
        messages.error(request, "You do not have permission to add items to this quote.")
        return redirect("quote", quote_id=quote_id)

    rate_id = request.POST.get("rate")
    rate: Optional[Rate] = None
    if rate_id:
        try:
            rate = Rate.non_deleted().get(id=rate_id)

        except Rate.DoesNotExist:
            messages.error(request, "Selected rate does not exist.")
            return redirect("quote", quote_id=quote_id)

    defaults = (
        {
            "description": rate.quote_display(),
            "amount": rate.amount,
            "minimum_charge": rate.minimum_charge,
            "service_fee": rate.service_fee,
            "rate_type": get_rate_type(rate),
        }
        if rate
        else {}
    )

    item_form = QuoteItemForm(merge_post_data_with_defaults(request.POST, defaults))

    if item_form.is_valid():
        quote_item = item_form.save(commit=False)
        quote_item.quote = quote_obj
        quote_item.save()
        messages.success(request, "Item added to quote successfully!")

        # Redirect to the quote page with the rate type and item id selected
        url = reverse("quote", kwargs={"quote_id": quote_id})
        if rate_type_choice or item_id:
            query_params = {}
            if rate_type_choice:
                query_params["type"] = rate_type_choice
            if item_id:
                query_params["id"] = item_id
            url += "?" + urlencode(query_params)
        return redirect(url)
    else:
        quote_items = quote_obj.quoteitem_set.all()
        empty_form = QuoteItemForm()
        context = {
            "quote": quote_obj,
            "quote_items": quote_items,
            "rate_item_form": item_form if active_form_type == "rate" else empty_form,
            "selected_rate": rate,
            "custom_item_form": item_form if active_form_type == "custom" else empty_form,
            "submitted": quote_obj.status == Quote.Status.PENDING_APPROVAL,
            "published": quote_obj.status == Quote.Status.PUBLISHED,
            "draft": quote_obj.status == Quote.Status.DRAFT,
            "updated": quote_obj.updated_date != quote_obj.created_date,
            "users": User.objects.filter(is_active=True).exclude(id=user.id),
            "projects": Project.objects.all(),
            "form": form,
            "active_quote_item_tab": active_form_type,
            "show_send_emails_button": show_send_emails_button,
        }
        rate_search_context = get_rate_search_context(rate_type_choice, item_id)
        context.update(rate_search_context)
        context.update(permissions)
        return render(request, "quotes/quote.html", context)


def get_rate_type(rate):
    if rate.is_hourly_rate():
        if rate.daily:
            return QuoteItem.AmountType.DAILY
        elif rate.flat:
            return QuoteItem.AmountType.FLAT
        else:
            return QuoteItem.AmountType.HOURLY
    else:
        return QuoteItem.AmountType.NOT_APPLICABLE


def merge_post_data_with_defaults(post, defaults):
    non_empty_post_data = {k: v for k, v in post.items() if v and v.strip() != ""}
    return {**defaults, **non_empty_post_data}


@login_required
@user_passes_test(can_create_quotes)
@require_http_methods(["POST"])
def delete_quote_item(request, item_id):
    quote_item = get_object_or_404(QuoteItem, id=item_id)
    permissions = quote_item.quote.permissions(request.user)
    if not permissions["can_edit"]:
        messages.error(request, "You do not have permission to delete items from this quote.")
        return redirect("quote", quote_id=quote_item.quote.id)

    quote_id = quote_item.quote.id
    quote_item.delete()
    messages.success(request, "Item removed from quote successfully!")
    return redirect("quote", quote_id=quote_id)


@login_required
@require_http_methods(["POST"])
def edit_quote_item(request, item_id):
    quote_item = get_object_or_404(QuoteItem, id=item_id)
    user = request.user
    permissions = quote_item.quote.permissions(user)

    if not permissions["can_edit"]:
        messages.error(request, "You do not have permission to edit items in this quote.")
        return redirect("quote", quote_id=quote_item.quote.id)

    # Update the quote item with form data
    item_form = QuoteItemForm(request.POST, instance=quote_item)

    if item_form.is_valid():
        item_form.save()
        messages.success(request, "Item updated successfully!")
        return redirect("quote", quote_id=quote_item.quote.id)
    else:
        quote_obj = quote_item.quote
        quote_items = quote_obj.quoteitem_set.all()
        form = EditQuoteForm(instance=quote_obj)
        show_send_emails_button = meets_send_email_prerequisites(quote_obj) and permissions["can_send_emails"]
        active_quote_item_tab = "rate"

        context = {
            "quote": quote_obj,
            "quote_items": quote_items,
            "submitted": quote_obj.status == Quote.Status.PENDING_APPROVAL,
            "published": quote_obj.status == Quote.Status.PUBLISHED,
            "draft": quote_obj.status == Quote.Status.DRAFT,
            "updated": quote_obj.updated_date != quote_obj.created_date,
            "users": User.objects.filter(is_active=True).exclude(id=user.id),
            "projects": Project.objects.all(),
            "form": form,
            "active_quote_item_tab": active_quote_item_tab,
            "show_send_emails_button": show_send_emails_button,
            "edit_item_form": item_form,
            "edit_item_id": item_id,
        }
        rate_search_context = get_rate_search_context(None, None)
        context.update(rate_search_context)
        context.update(permissions)

        return render(request, "quotes/quote.html", context)


@login_required
@user_passes_test(can_create_quotes)
@require_http_methods(["POST"])
def add_tax(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    permissions = quote_obj.permissions(user)
    if not permissions["can_edit"]:
        messages.error(request, "You do not have permission to add tax to this quote.")
        return redirect("quote", quote_id=quote_id)

    if quote_obj.configuration.tax is None:
        messages.error(request, "This quote configuration does not have a tax.")
        return redirect("quote", quote_id=quote_id)

    quote_obj.add_tax = True
    quote_obj.save()
    messages.success(request, "Tax added to quote successfully!")
    return redirect("quote", quote_id=quote_id)


@login_required
@user_passes_test(can_create_quotes)
@require_http_methods(["POST"])
def remove_tax(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    permissions = quote_obj.permissions(user)
    if not permissions["can_edit"]:
        messages.error(request, "You do not have permission to remove tax from this quote.")
        return redirect("quote", quote_id=quote_id)

    quote_obj.add_tax = False
    quote_obj.save()
    messages.success(request, "Tax removed from quote successfully!")
    return redirect("quote", quote_id=quote_id)


def get_rate_search_context(rate_type_choice, item_id):
    tool = get_object_or_404(Tool, pk=item_id) if item_id and rate_type_choice == RateType.Type.TOOL else None
    area = get_object_or_404(Area, pk=item_id) if item_id and rate_type_choice == RateType.Type.AREA else None
    consumable = (
        get_object_or_404(Consumable, pk=item_id) if item_id and rate_type_choice == RateType.Type.CONSUMABLE else None
    )
    tools = list(Tool.objects.all())
    tools.sort(key=lambda x: (x.category, x.name))
    rate_types = get_rate_types(rate_type_choice)
    rates = []
    search_selection = None

    if rate_type_choice:
        if rate_type_choice in [RateType.Type.TOOL, RateType.Type.AREA, RateType.Type.CONSUMABLE]:
            if tool or area or consumable:
                rates = Rate.non_deleted().filter(type__in=rate_types)
                if tool:
                    rates = rates.filter(tool=tool)
                    search_selection = tool.__str__()
                elif area:
                    rates = rates.filter(area=area)
                    search_selection = area.__str__()
                elif consumable:
                    rates = rates.filter(consumable=consumable)
                    search_selection = consumable.__str__()
        else:
            rates = Rate.non_deleted().filter(type__in=rate_types)
            search_selection = rate_types.first().__str__().lower()

    current_and_future_rates = flatten(RateHistory(rates, datetime.today().date()).current_and_future_rates().values())
    current_and_future_rates.sort(key=lambda r: (r.type_id, r.category_id))

    return {
        "rate_types": rate_types,
        "rate_type_choice": rate_type_choice,
        "item_id": item_id,
        "item": tool or area or consumable,
        "rate_type_choices": get_rate_type_choices(),
        "tools": tools,
        "areas": Area.objects.filter(area_children_set__isnull=True),
        "consumables": Consumable.objects.all().order_by("category", "name"),
        "rates": current_and_future_rates,
        "search_selection": search_selection,
        "found_count": len(current_and_future_rates),
    }


@login_required
@user_passes_test(can_access_quotes)
@require_http_methods(["POST"])
def send_quote(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    permissions = quote_obj.permissions(user)

    if not permissions["can_send_emails"]:
        messages.error(request, "You do not have permission to send this quote.")
        return redirect("quote", quote_id=quote_id)

    try:
        send_quote_email(quote_obj)
        messages.success(request, "Quote sent successfully!")
        quote_obj.last_emails_sent_date = timezone.now()
        quote_obj.save()
    except FailedToSendQuoteEmailException:
        messages.error(request, "Failed to send quote email.")

    return redirect("quote", quote_id=quote_id)


@login_required
@user_passes_test(can_access_quotes)
@require_http_methods(["POST"])
def render_quote(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    permissions = quote_obj.permissions(user)

    if not permissions["can_send_emails"]:
        messages.error(request, "You do not have permission to render this quote.")
        return redirect("quote", quote_id=quote_id)

    try:
        quote_obj.render_and_save()
        messages.success(request, "Quote rendered successfully!")
    except QuoteGenerationException as e:
        messages.error(request, "Failed to render quote file.")

    return redirect("quote", quote_id=quote_id)


@require_GET
def public_view_quote(request, quote_id):
    try:
        quote = Quote.objects.filter(id=quote_id).first()
        token = request.GET.get("token")
        if not token or not quote.file_access_token or token != quote.file_access_token or not quote.file:
            return HttpResponse("Invalid Link", content_type="text/plain")
        if quote.is_expired:
            return HttpResponse("Link Expired", content_type="text/plain")
        return FileResponse(quote.file, as_attachment=False, filename=quote.file.name.split("/")[-1])
    except Exception:
        quotes_logger.exception(f"Failed to serve quote file for quote {quote_id}")
        return HttpResponse("Invalid Link", content_type="text/plain")


@login_required
@user_passes_test(can_access_quotes)
@require_http_methods(["POST"])
def update_quote_status(request, quote_id):
    quote_obj = get_object_or_404(Quote, id=quote_id)
    user = request.user
    permissions = quote_obj.permissions(user)
    action = next(
        iter(
            [
                state
                for state in ["approve_quote", "publish_quote", "submit_quote", "deny_quote"]
                if state in request.POST
            ]
        )
    )

    if action:
        if action == "approve_quote":

            if not permissions["can_approve"]:
                messages.error(request, "User does not have permission to approve this quote.")
                return redirect("quote", quote_id=quote_id)

            if not quote_obj.is_pending_approval:
                messages.error(request, "Only quotes pending approval can be approved.")
                return redirect("quote", quote_id=quote_id)

            quote_obj.publish(user)
            send_status_update_email(quote_obj)
            delete_approval_request_notification(quote_obj)
            messages.success(request, "Quote published successfully!")

        elif action == "deny_quote":

            if not permissions["can_approve"]:
                messages.error(request, "User does not have permission to deny this quote.")

            if not quote_obj.is_pending_approval:
                messages.error(request, "Only quotes pending approval can be denied.")

            quote_obj.deny()
            send_status_update_email(quote_obj)
            delete_approval_request_notification(quote_obj)
            messages.success(request, "Quote denied successfully!")

        elif action == "submit_quote":

            if not permissions["can_edit"]:
                messages.error(request, "User does not have permission to submit this quote for approval.")

            if not quote_obj.is_draft:
                messages.error(request, "Only draft quotes can be submitted for approval.")

            if not permissions["requires_approval"]:
                messages.error(request, "This quote does not require approval.")

            quote_obj.submit_for_approval()
            send_approval_request_email(quote_obj)
            create_approval_request_notification(quote_obj)
            messages.success(request, "Quote submitted for review successfully!")

        elif action == "publish_quote":

            if not permissions["can_edit"]:
                messages.error(request, "User does not have permission to publish this quote.")

            if permissions["requires_approval"]:
                messages.error(request, "This quote requires approval before it can be published.")

            if not quote_obj.is_draft:
                messages.error(request, "Only draft quotes can be published.")

            quote_obj.publish(user)
            delete_approval_request_notification(quote_obj)
            messages.success(request, "Quote published successfully!")

    return redirect("quote", quote_id=quote_id)


def send_status_update_email(quote):
    try:
        subject = f"Quote Status Update: {quote.name} - {quote.get_status_display()}"
        quote_status_update_email_template = get_media_file_contents("quote_status_update_email.html")
        if quote_status_update_email_template:
            content = render_email_template(quote_status_update_email_template, {"quote": quote})
            quote.creator.email_user(
                subject=subject, message=content, from_email=None, email_category=QUOTE_EMAIL_CATEGORY
            )
    except Exception as e:
        quotes_logger.exception(f"Failed to send status update email for quote {quote.name} ({quote.id})", e)


def send_approval_request_email(quote):
    try:
        subject = f"Quote Approval Request: {quote.name}"
        quote_approval_request_email_template = get_media_file_contents("quote_approval_request_email.html")
        if quote_approval_request_email_template:
            content = render_email_template(quote_approval_request_email_template, {"quote": quote})
            recipients = quote.configuration.get_all_reviewers()
            for recipient in recipients:
                recipient.email_user(
                    subject=subject, message=content, from_email=None, email_category=QUOTE_EMAIL_CATEGORY
                )
    except Exception as e:
        quotes_logger.exception(f"Failed to send approval request email for quote {quote.name} ({quote.id})", e)


def send_quote_email(quote):
    subject = f"Quote: {quote.name}"
    quote_email_template = get_media_file_contents("quote_email.html")
    if quote_email_template and quote.file_access_token and quote.file:
        content = render_email_template(quote_email_template, {"quote": quote, "quote_url": quote.get_access_url()})
        recipients = quote.get_all_recipients()
        try:
            send_mail(
                subject=subject,
                content=content,
                from_email=quote.creator.email,
                bcc=recipients,
                cc=quote.configuration.email_cc,
                email_category=QUOTE_EMAIL_CATEGORY,
                fail_silently=False,
            )
        except Exception:
            quotes_logger.exception(f"Failed to send quote email for quote {quote.name} ({quote.id}) to {recipients}")
            raise FailedToSendQuoteEmailException(quote)



# Generated by Django 2.2.24 on 2021-09-02 19:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0006_version_1_7_1"),
    ]

    operations = [
        migrations.AddField(
            model_name="projectbillingdetails",
            name="no_tax",
            field=models.BooleanField(default=False, help_text="Check this box if this project is tax exempt."),
        ),
    ]



from datetime import datetime
from itertools import groupby
from typing import Dict, Iterable, List

from NEMO.models import Consumable, Tool, User
from NEMO.rates import Rates
from NEMO.utilities import distinct_qs_value_list, format_datetime
from django.conf import settings
from django.db.models import Q
from django.utils.formats import number_format

from NEMO_billing.invoices.models import InvoiceConfiguration
from NEMO_billing.invoices.processors import get_rate_with_currency
from NEMO_billing.invoices.utilities import flatten
from NEMO_billing.rates.models import Rate, RateCategory, RateType
from NEMO_billing.rates.utilities import RateHistory, get_rate_history


class DatabaseRates(Rates):

    expand_rates_table = False

    def __init__(self):
        self.currency = getattr(settings, "RATE_CURRENCY", "$")
        self.configuration = InvoiceConfiguration.first_or_default()

    def load_rates(self, force_reload=False):
        super().load_rates(force_reload=force_reload)

    def get_consumable_rates(self, consumables: List[Consumable], user: User = None) -> Dict[str, str]:
        from NEMO_billing.rates.customization import BillingRatesCustomization

        if BillingRatesCustomization.get_bool("rates_hide_consumable_rates"):
            return {}

        # Initialize consumable rates loading before looking up individual rates
        self.configuration.init_rate_help(Q(type__in=RateType.objects.filter(type=RateType.Type.CONSUMABLE)))

        return {consumable.name: self.get_consumable_rate(consumable, user) for consumable in consumables}

    def get_consumable_rate(self, consumable: Consumable, user: User = None) -> str:
        from NEMO_billing.rates.customization import BillingRatesCustomization

        today = datetime.today().date()

        if BillingRatesCustomization.get_bool("rates_hide_consumable_rates"):
            return ""

        consumable_rates = get_rate_history(
            self.configuration.rates_types.get(RateType.Type.CONSUMABLE).id,
            self.configuration.all_rates,
            consumable_id=consumable.id,
            as_of_date=today,
        ).current_and_future_rates()

        consumable_rates = flatten(consumable_rates.values())

        cons_rate_display_list = []
        for cons_rate in consumable_rates:
            cons_rate_display = self.consumable_rate_display(cons_rate)
            if cons_rate.effective_date and cons_rate.effective_date > today:
                cons_rate_display += (
                    f" <b>effective {format_datetime(cons_rate.effective_date, 'SHORT_DATE_FORMAT')}</b>"
                )
            cons_rate_display_list.append(cons_rate_display)

        return ", ".join(cons_rate_display_list) if cons_rate_display_list else ""

    def get_tool_rates(self, tools: List[Tool], user: User = None) -> Dict[str, str]:
        from NEMO_billing.rates.customization import BillingRatesCustomization

        if BillingRatesCustomization.get_bool("rates_hide_table"):
            return {}
        return super().get_tool_rates(tools, user)

    def get_tool_rate(self, tool: Tool, user: User = None) -> str:
        from NEMO_billing.rates.customization import BillingRatesCustomization

        if BillingRatesCustomization.get_bool("rates_hide_table"):
            return ""
        all_tool_rates = Rate.non_deleted().filter(tool=tool).order_by("type", "category")
        if not all_tool_rates:
            return ""
        show_all_categories = BillingRatesCustomization.get_bool("rates_show_all_categories")
        if user and not user.is_any_part_of_staff and not show_all_categories and RateCategory.objects.exists():
            user_categories = distinct_qs_value_list(
                user.active_projects().filter(projectbillingdetails__category__isnull=False),
                "projectbillingdetails__category",
            )
            if not user_categories:
                return ""
            all_tool_rates = all_tool_rates.filter(category__in=user_categories)
        all_tool_rates = list(all_tool_rates)
        rate_categories = sorted(set([rate.category.name for rate in all_tool_rates]))
        list_by_type = groupby(all_tool_rates, key=lambda x: x.type)
        html_rate = f'<div class="media"><a onclick="toggle_details(this)" class="pointer collapsed" data-toggle="collapse" data-target="#rates_details"><span class="glyphicon glyphicon-list-alt pull-left notification-icon primary-highlight"></span><span class="glyphicon pull-left chevron glyphicon-chevron-{"down" if self.get_expand_rates_table() else "right"}"></span></a>'
        html_rate += f'<div class="media-body"><span class="media-heading">Rates</span><div id="rates_details" class="collapse {"in" if self.get_expand_rates_table() else ""}"><table class="table table-bordered table-hover thead-light" style="width: auto !important; margin-bottom: 0">'
        if rate_categories:
            html_rate += '<tr><th class="text-center"></th><th class="text-center">'
            html_rate += '</th><th class="text-center">'.join(rate_categories)
            html_rate += "</tr>"
        for rate_type, tool_rates in list_by_type:
            html_rate += (
                f'<tr><th class="text-center" style="vertical-align: middle">{rate_type.get_type_display()}</th>'
            )
            if not rate_type.category_specific or not RateCategory.objects.exists():
                html_rate += f'<td class="text-center" style="vertical-align: middle" colspan="{len(rate_categories)}">{self.tool_rate_display_with_details(tool_rates)}</td>'
            else:
                for rate_category in rate_categories:
                    tool_rate_category = [
                        rate
                        for rate in all_tool_rates
                        if rate.type == rate_type and rate.category.name == rate_category
                    ]
                    html_rate += f'<td class="text-center" style="vertical-align: middle">{self.tool_rate_display_with_details(tool_rate_category)}</td>'
        html_rate += "</tr></table></div></div></div>"
        return html_rate

    def tool_rate_display_with_details(self, rates: Iterable[Rate]):
        today = datetime.today().date()

        display_rates = flatten(RateHistory(rates, today).current_and_future_rates().values())

        return "<br>".join(
            [
                f"{get_rate_with_currency(self.configuration, rate.display_rate())}{' (' + rate.time.name + ')' if rate.time else ''}{' <b>effective '+format_datetime(rate.effective_date, 'SHORT_DATE_FORMAT') + '</b>' if rate.effective_date and rate.effective_date > today else ''}"
                for rate in display_rates
            ]
        )

    def consumable_rate_display(self, rate: Rate) -> str:
        return f"<b>{self.display_amount(rate.amount)}</b>"

    def display_amount(self, amount):
        return f"{self.currency}{number_format(amount, decimal_pos=2)}"



from typing import List

from NEMO_billing.exceptions import BillingException
from NEMO_billing.quotes.models import Quote


class QuoteGenerationException(BillingException):
    def __init__(self, quote: Quote, renderer, e: Exception):
        self.quote = quote
        self.renderer = renderer
        msg = f"Error rendering quote {quote.name} ({quote.id}) with {renderer.__class__.__name__}: {str(e)}."
        super().__init__(msg)


class FailedToSendQuoteEmailException(BillingException):
    def __init__(self, quote: Quote):
        self.quote = quote
        msg = f"Error sending quote {quote.name} ({quote.id}) email."
        super().__init__(msg)



from django.forms import ModelForm

from NEMO_billing.quotes.models import Quote, QuoteItem


class CreateQuoteForm(ModelForm):
    class Meta:
        model = Quote
        fields = ["name", "configuration"]


class EditQuoteForm(ModelForm):
    class Meta:
        model = Quote
        fields = ["name", "project", "users", "emails"]


class QuoteItemForm(ModelForm):
    class Meta:
        model = QuoteItem
        exclude = ["quote"]



# Generated by Django 2.2.24 on 2021-09-14 14:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0007_version_1_8_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoicepayment",
            name="note",
            field=models.CharField(blank=True, help_text="Payment note", max_length=255, null=True),
        ),
    ]



# Generated by Django 2.2.24 on 2021-07-30 17:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0004_version_1_6_0"),
    ]

    operations = [
        migrations.AlterField(
            model_name="projectbillingdetails",
            name="project",
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to="NEMO.Project"),
        ),
    ]
